unit U_gener;

interface

uses ExtCtrls, SysUtils;

const MAXLAENGE = 50;                                           { maximale Lnge des Gencodes     }
      MAXANZAHL = 200;                                          { maximale Gre einer Generation }

type TIndividuum = array[1..MAXLAENGE] of byte;                 { Gencode eines Individuums       }
     TIndividuen = array[1..MAXANZAHL] of TIndividuum;          { Gencode einer Generation        }
     TGeneration = class                                        { Eine Generation besteht aus:    }
       private
         BisherBestesIndividuum: TIndividuum;                   { Nur zu Anzeigezwecken           }
         BisherBesteFitness: longint;

         Individuen: TIndividuen;                               {   dem Gencode der Individuen    }
         PopulationsGroesse: longint;                           {   einer Grenbeschrnkung      }
         CodeLaenge: longint;                                   {   einer Gencode-Lngenbeschrnk.}

         procedure SetIndividuum(Nummer: longint; Individuum: TIndividuum);
            { Setzt das Individuum mit dem Index "Nummer" neu                                     }
         function ZufaelligesIndividuum: TIndividuum;
            { Liefert ein zuflliges Individuum nach der Roulette-Methode                         }

       protected
         Mutationsrate: longint;  { in Prozent }                {   einer vorgegebenen Mutationsr.}
         Kreuzungsrate: longint;  { in Prozent }                {   und einer Kreuzungsrate       }

         procedure Mutation(var Individuum: TIndividuum); virtual;
            { Mutiert das mitgegebene Individuum. Diese Methode kann berschrieben werden.        }
         procedure Kreuzung(var Individuum1, Individuum2: TIndividuum); virtual;
            { Kreuzt zwei Individuen. Diese Methode kann berschrieben werden.                    }
         function BestesIndividuum: TIndividuum;
            { Liefert das beste Individuum der aktuellen Generation.                              }
       public
         constructor Create(Anzahl: longint; Laenge: longint; Mutation: longint; Kreuzung: longint);
            { Erzeugt eine neue Generation mit zuflliger Belegung                                }
              { Anzahl: Anzahl der Individuen in einer Generation                                 }
              { Laenge: bentigte Lnge des Gencodes eines Individuums                            }
              { Mutation: Mutationsrate in Prozent                                                }
              { Kreuzung: Kreuzungsrate in Prozent                                                }
         function Fitness(Individuum: TIndividuum): longint; virtual; abstract;
            { Diese Funktion muss (!) von jeder Anwendung selbst implementiert werden, da die     }
            { Fitness einer Genration vom konkreten Anwendungszusammenhang abhngig ist.          }
         procedure NaechsteGenerationErzeugen_BesteBleibenErhalten;
            { Wechselt die Generation Schritt fr Schritt, d. h., jeweils zwei nach der Roulette- }
            { Methode selektierten, mutierten und gekreuzten Kinder ersetzen die zwei             }
            { schlechtesten Individuen der Generation.                                            }
         procedure NaechsteGenerationErzeugen_Komplett;
            { Wechselt die aktuelle Generation. Alle Eltern sterben, die Kinder entstehen durch   }
            { Selektion nach Roulette-Methode, Kreuzung und Mutation.                             }
         function Individuum(Nummer: longint): TIndividuum;
            { Liefert das Individuum mit dem Index "Nummer"                                       }
         procedure Anzeigen(PB: TPaintbox);
            { Mglichkeit der Anzeige der Generation als Streudiagramm (Index -> Fitness          }
            { Das bisher beste Individuum wird durch einen Balken gekennzeichnet.                 }
     end;

implementation


function TGeneration.ZufaelligesIndividuum: TIndividuum;
{ zufllige Auswahl nach dem Roulette-Modell, d. h. die Wahrscheinlichkeit, ausgewhlt zu werden
  ist        Fitness
       p = --------------
           Gesamt-Fitness
}
var FitnessGesamt: longint;
    i: integer;
    Zufallszahl: longint;
    Nummer: integer;
begin
  FitnessGesamt:= 0;                                                 { Gesamt-Fitness feststellen }
  for i:= 1 to Populationsgroesse
  do FitnessGesamt:= FitnessGesamt+Fitness(Individuum(i));
  if FitnessGesamt=0                                                 { Gesamt-Fitness = 0, dann   }
  then result:= Individuum(1)                                        {    whle immer erstes Ind. }
  else begin
         Zufallszahl:= random(FitnessGesamt);                        { sonst ein zuflliges nach  }
         Nummer:= 0;                                                 { Roulette-Methode           }
         repeat
           inc(Nummer);
           Zufallszahl:= Zufallszahl-Fitness(Individuum(Nummer));
         until Zufallszahl<0;
         result:= Individuum(Nummer)
       end;
end;


constructor TGeneration.Create(Anzahl: longint; Laenge: longint; Mutation: longint; Kreuzung: longint);
{ Erzeugt und initialisiert eine neue Generation. Die Anfangsindividuen sind vom Kontext
  unabhngig, zufllig erstellt. Mutationsrate und Kreuzungsrate werden festgelegt.
}
var i, j: integer;
    Individuum: TIndividuum;
begin
  PopulationsGroesse:= Anzahl;
  CodeLaenge:= Laenge;
  Mutationsrate:= Mutation;
  Kreuzungsrate:= Kreuzung;

  for i:= 1 to Populationsgroesse                                    { Zufllige Erstellung des   }
  do begin                                                           { Gencodes der Individuen    }
       for j:= 1 to CodeLaenge
       do Individuum[j]:= random(2);
       SetIndividuum(i,Individuum);
     end;
  BisherBestesIndividuum:= BestesIndividuum;                         { Zu Anzeigezwecken bentigt }
  BisherBesteFitness:= Fitness(BisherBestesIndividuum);
end;

function TGeneration.Individuum(Nummer: longint): TIndividuum;
{ Sogenannter "Getter"
}
begin
  result:= Individuen[Nummer];
end;

procedure TGeneration.SetIndividuum(Nummer: longint; Individuum: TIndividuum);
{ Sogenannter "Setter"
}
begin
  Individuen[Nummer]:= Individuum;
end;

function TGeneration.BestesIndividuum: TIndividuum;
{ Liefert das beste Individuum der aktuellen Generation.
}
var i: integer;
    maxFitness: longint;
    BestesIndividuum: TIndividuum;
begin
  maxFitness:= Fitness(Individuum(1));                               { Beste Fitness initialisieren }
  BestesIndividuum:= Individuum(1);
  for i:= 2 to Populationsgroesse                                    { Fr alle Individuen testen:  }
  do if Fitness(Individuum(i))>maxFitness                            {   gibt es eine bessere?      }
     then begin
            maxFitness:= Fitness(Individuum(i));                     {   dann neu setzen            }
            BestesIndividuum:= Individuum(i);
          end;
  result:= BestesIndividuum;
end;

procedure TGeneration.Mutation(var Individuum: TIndividuum);
{ Die Standardmutation kippt ein zuflliges Bit im Individuum.
  Allerdings kann es in anderen Zusammenhnge eine sinnvollere Mutation geben.
  Wird der Gencode z. B. als Integer-Zahl aufgefasst, so knnte eine Mutation
  darin bestehen, die Zahl um eine Zufallszahl zu vergrern oder zu verkleinern.
  Deshalb ist diese Methode berschreibbar"
}
var Stelle: integer;
begin
  if random(100)+1 <= Mutationsrate                                  { Mutationsrate in Prozent  }
  then begin
         Stelle:= random(CodeLaenge)+1;                              { Stelle zufllig aussuchen }
         Individuum[Stelle]:= 1-Individuum[Stelle];                  {  und Bit kippen           }
       end;
end;

procedure TGeneration.Kreuzung(var Individuum1, Individuum2: TIndividuum);
{ Die Standardkreuzung vertauscht ab einer zufllig bestimmten Stelle den Gencode
  zweier Individuen. Allerdings kann in anderen Zusammenhngen eine andere
  Kreuzung sinnvoller sein. Soll z. B. das Maximum einer Funktion f(x) gefunden werden
  und der Gencode gibt die x-Koordinate an, so sollte eine Kreuzung darin bestehen,
  dass zwei neue Individuen in der Mitte beider Eltern erzeugt werden.}
var Hilfsbit: byte;
    i: integer;
    Stelle: integer;
begin
  if random(100)+1 <= Kreuzungsrate                                  { Kreuzungsrate in Prozent  }
  then begin
         Stelle:= random(CodeLaenge)+1;                              { Stelle zufllig aussuchen }
         for i:= 1 to Stelle
         do begin                                                    {   und alle Bits bis dort  }
              Hilfsbit:= Individuum1[i];                             {   hin vertauschen         }
              Individuum1[i]:= Individuum2[i];
              Individuum2[i]:= Hilfsbit;
            end;
       end;
end;


procedure TGeneration.NaechsteGenerationErzeugen_BesteBleibenErhalten;
{ Wechselt die Generation Schritt fr Schritt, d. h., jeweils zwei nach der Roulette-
  Methode selektierten, mutierten und gekreuzten Kinder ersetzen die zwei schlechtesten Individuen
  der Generation.
}
var i,j: integer;
    NeueIndividuen: TIndividuen;
    Eltern1, Eltern2, Kind1, Kind2: TIndividuum;

    function schlechtestesIndividuum: integer;
    var min: integer;
        i: integer;
    begin
      min:= 1;
      for i:= 1 to Populationsgroesse
      do if Fitness(Individuen[i])<Fitness(Individuen[min])
         then min:= i;
      result:= min;
    end;


begin
  for i:= 0 to  Populationsgroesse div 2 - 1
  do begin
       Eltern1:= ZufaelligesIndividuum;                       { 2 Elternteile aussuchen                 }
       Eltern2:= ZufaelligesIndividuum;

       Kind1:= Eltern1;                                       { Kinder erzeugen (Kopie der Eltern)      }
       Kind2:= Eltern2;

       Kreuzung(Kind1, Kind2);                                { Gene eventuell kreuzen (Kreuzungsrate)  }

       Mutation(Kind1);                                       { Gene Eventuell mutieren (Mutationsrate) }
       Mutation(Kind2);

       Individuen[schlechtestesIndividuum]:= Kind1;           { Kinder in neue Generation aufnehmen     }
       Individuen[schlechtestesIndividuum]:= Kind2;
     end;

  if Fitness(BestesIndividuum)>BisherBesteFitness             { Zu Anzeigezwecken bentigt              }
  then begin
         BisherBestesIndividuum:= BestesIndividuum;
         BisherBesteFitness:= Fitness(BisherBestesIndividuum);
       end;

end;


procedure TGeneration.NaechsteGenerationErzeugen_Komplett;
{ Wechselt die Generation vollstndig, d. h., alle Eltern sterben und die nach der Roulette-
  Methode selektierten, mutierten und gekreuzten Kinder bernehmen die "Macht"
}
var i,j: integer;
    NeueIndividuen: TIndividuen;
    Eltern1, Eltern2, Kind1, Kind2: TIndividuum;
begin
  for i:= 1 to MAXANZAHL do for j:= 1 to MAXLAENGE do NeueIndividuen[i][j]:= 0;
  for i:= 0 to Populationsgroesse div 2 - 1
  do begin
       Eltern1:= ZufaelligesIndividuum;                       { 2 Elternteile aussuchen                 }
       Eltern2:= ZufaelligesIndividuum;

       Kind1:= Eltern1;                                       { Kinder erzeugen (Kopie der Eltern)      }
       Kind2:= Eltern2;

       Kreuzung(Kind1, Kind2);                                { Gene eventuell kreuzen (Kreuzungsrate)  }

       Mutation(Kind1);                                       { Gene Eventuell mutieren (Mutationsrate) }
       Mutation(Kind2);

       NeueIndividuen[i*2+1]:= Kind1;                         { Kinder in neue Generation aufnehmen     }
       NeueIndividuen[i*2+2]:= Kind2;
     end;
  Individuen:= NeueIndividuen;

  if Fitness(BestesIndividuum)>BisherBesteFitness             { Zu Anzeigezwecken bentigt              }
  then begin
         BisherBestesIndividuum:= BestesIndividuum;
         BisherBesteFitness:= Fitness(BisherBestesIndividuum);
       end;

end;

procedure TGeneration.Anzeigen(PB: TPaintbox);
{ Zeigt die aktuelle Generation als Streudiagramm Index -> Fitness an.
}
var i: integer;
    Distanzx, Distanzy: single;
begin
  PB.Refresh;
  Distanzx:= PB.Width / (Populationsgroesse+1);                { Abstand zwischen Punkten in x-Richtung }
  Distanzy:= (PB.Height-20) / (BisherBesteFitness+1);          { Abstand zwischen Punkten in y-Richtung }

  for i:= 1 to Populationsgroesse                              { Alle Punkte (Index|Fitness) zeichnen   }
  do begin
       PB.Canvas.Rectangle( trunc(i*Distanzx), trunc(PB.Height-10-Fitness(Individuum(i))*Distanzy),
                            trunc(i*Distanzx+4), trunc(PB.Height-10-Fitness(Individuum(i))*Distanzy+4));
       PB.Canvas.MoveTo(0,trunc(PB.Height-10-BisherBesteFitness*Distanzy));
       PB.Canvas.LineTo(PB.Width,trunc(PB.Height-10-BisherBesteFitness*Distanzy));
     end;

                                                               { Balken bei bisher bester Fitness setzen}
  PB.Canvas.TextOut(0,trunc(PB.Height-10-BisherBesteFitness*Distanzy-PB.Canvas.TextHeight('X') / 2),
                    IntToStr(BisherBesteFitness));

end;


end.
