unit U_io;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, U_Gener, U_Berge;

type
  TBergGeneration = class (TGeneration)
    function Fitness(Individuum: TIndividuum): longint; override;            { Fitness proportional zum Radius }
    procedure Mutation(var Individuum: TIndividuum);    override;            { neue Mutations-Vorschrift       }
    procedure Kreuzung(var Individuum1, Individuum2: TIndividuum); override; { neue Kreuzungs-Vorschrift       }
    function IndividuumToPunkt(Individuum: TIndividuum): TPoint;             { Umwandlung Binrcode in (x|y)   }
    function PunktToIndividuum(Mittelpunkt: TPoint): TIndividuum;            { Uwandlung (x|y) in Binrcode   }
  end;


  TF_Berge = class(TForm)
    PB_Gebirge: TPaintBox;
    B_Neustart: TButton;
    B_Naechste: TButton;
    B_Erste: TButton;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    E_Anzahl: TEdit;
    E_Mutation: TEdit;
    E_Kreuzung: TEdit;
    GroupBox2: TGroupBox;
    PB_GenAlg: TPaintBox;
    E_Nr: TEdit;
    B_Best: TButton;
    E_Best: TEdit;
    Label4: TLabel;
    Timer1: TTimer;
    L_Hoehe: TLabel;
    procedure B_ErsteClick(Sender: TObject);
    procedure B_NeustartClick(Sender: TObject);
    procedure B_NaechsteClick(Sender: TObject);
    procedure B_BestClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure PB_GebirgeMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
    Gebirge: TGebirge;
    BergGeneration: TBergGeneration;
    procedure PunktAnzeigen(Punkt: TPoint; Wert: integer; c: TColor);
  public
    { Public declarations }
  end;

var
  F_Berge: TF_Berge;

implementation

{$R *.DFM}


function TBergGeneration.IndividuumToPunkt(Individuum: TIndividuum): TPoint;
{ Ein Individuum mit 16 Bit ist ein Punkt, wobei die ersten 8 Bit der x-KO des Mittelpunkts,
  die zweiten 8 Bit der y-KO des Mittelpunktes sind. Der Radius ergibt sich ber die Fitness }
var i: integer;
    Mittelpunkt: TPoint;
begin
  Mittelpunkt.x:= 0;
  for i:= 1 to 8                                                    { Umwandlung der ersten 8 Bit in eine Int-Zahl }
  do Mittelpunkt.x:= 2*Mittelpunkt.x+ Individuum[i];
  Mittelpunkt.y:= 0;
  for i:= 9 to 16                                                   { Umwandlung der zweiten 8 Bit in eine Int-Zahl }
  do Mittelpunkt.y:= 2*Mittelpunkt.y+ Individuum[i];
  Result:= Mittelpunkt;
end;

function TBergGeneration.PunktToIndividuum(Mittelpunkt: TPoint): TIndividuum;
{ Ein Mittelpunkt wird in den Gencode eines Individuums berfhrt. Die x-KO entsprechen
  8 Bit, sie y-KO entsprechen den zweiten 8 Bit.
}
var i: integer;
    Individuum: TIndividuum;
begin
  for i:= 8 downto 1
  do begin
       Individuum[i]:= Mittelpunkt.x mod 2;
       Mittelpunkt.x:= Mittelpunkt.x div 2;
     end;
  for i:= 16 downto 9
  do begin
       Individuum[i]:= Mittelpunkt.y mod 2;
       Mittelpunkt.y:= Mittelpunkt.y div 2;
     end;
  Result:= Individuum;
end;


function TBergGeneration.Fitness(Individuum: TIndividuum): longint;
{  Die Fitness eines Individuums definiert durch die Hhe des Punktes im Gebirge.
}
var Punkt: TPoint;
begin
  Punkt:=IndividuumToPunkt(Individuum);
  if (Punkt.x=0) or (Punkt.y=0)
  then Fitness:= 0
  else Fitness:= F_Berge.Gebirge.Punkte[Punkt.x,Punkt.y];
end;

procedure TBergGeneration.Mutation(var Individuum: TIndividuum);
{ Die Koordinaten des Punktes werden evtl. leicht verndert.
}
var Mittelpunkt: TPoint;
begin
  if random(100)+1 <= Mutationsrate
  then begin
         Mittelpunkt:= IndividuumToPunkt(Individuum);
         Mittelpunkt.x:= Mittelpunkt.x+random(21)-10;
         Mittelpunkt.y:= Mittelpunkt.y+random(21)-10;
         Individuum:= PunktToIndividuum(Mittelpunkt);
       end;
end;

procedure TBergGeneration.Kreuzung(var Individuum1, Individuum2: TIndividuum);
{ Punkt zufllig zwischen Punkten der beiden Individuen.
}
var
    Mittelpunkt1, Mittelpunkt2: TPoint;
    NeuMittelpunkt1, NeuMittelpunkt2: TPoint;

begin
  if random(100)+1 <= Kreuzungsrate
  then begin
         Mittelpunkt1:= IndividuumToPunkt(Individuum1);
         Mittelpunkt2:= IndividuumToPunkt(Individuum2);

         if Mittelpunkt1.x>Mittelpunkt2.x
         then begin
                NeuMittelpunkt1.x:= random(Mittelpunkt1.x-Mittelpunkt2.x+1)+Mittelpunkt2.x;
                NeuMittelpunkt2.x:= random(Mittelpunkt1.x-Mittelpunkt2.x+1)+Mittelpunkt2.x
              end
         else begin
                NeuMittelpunkt1.x:= random(Mittelpunkt2.x-Mittelpunkt1.x+1)+Mittelpunkt1.x;
                NeuMittelpunkt2.x:= random(Mittelpunkt2.x-Mittelpunkt1.x+1)+Mittelpunkt1.x;
              end;
         if Mittelpunkt1.y>Mittelpunkt2.y
         then begin
                NeuMittelpunkt1.y:= random(Mittelpunkt1.y-Mittelpunkt2.y+1)+Mittelpunkt2.y;
                NeuMittelpunkt2.y:= random(Mittelpunkt1.y-Mittelpunkt2.y+1)+Mittelpunkt2.y
              end
         else begin
                NeuMittelpunkt1.y:= random(Mittelpunkt2.y-Mittelpunkt1.y+1)+Mittelpunkt1.y;
                NeuMittelpunkt2.y:= random(Mittelpunkt2.y-Mittelpunkt1.y+1)+Mittelpunkt1.y;
              end;

         Individuum1:= PunktToIndividuum(NeuMittelpunkt1);
         Individuum2:= PunktToIndividuum(NeuMittelpunkt2);
       end;
end;


procedure TF_Berge.PunktAnzeigen(Punkt: TPoint; Wert: integer; c: TColor);
{ Anzeige eines Punktes des Gebirges
}
begin
  PB_Gebirge.Canvas.Pen.Color:= c;
  PB_Gebirge.Canvas.Brush.Color:= c;
  PB_Gebirge.Canvas.Ellipse(Punkt.x-3,Punkt.y-3,Punkt.x+3,Punkt.y+3);
  if wert >= 0
  then begin
         PB_Gebirge.Canvas.Font.Color:= c;
         PB_Gebirge.Canvas.Brush.Color:= clwhite;
         PB_Gebirge.Canvas.TextOut(Punkt.x+2,Punkt.y+2,IntToStr(Wert));
       end;
end;


procedure TF_Berge.B_ErsteClick(Sender: TObject);
{ Erste Generation (Start) bilden und anzeigen.
}
var Punkt: TPoint;
    i: integer;
begin
  E_Nr.Text:= '0';
  B_Naechste.Enabled:= true;
  B_Naechste.SetFocus;
  BergGeneration.Free;
  BergGeneration:= TBergGeneration.Create(StrToInt(E_Anzahl.Text),16,StrToInt(E_Mutation.Text),StrToInt(E_Kreuzung.Text));
                                                                  { Erste Generation erzeugen (zufllig) }
  BergGeneration.Anzeigen(PB_GenAlg);

  Gebirge.Anzeigen;

  Punkt:= BergGeneration.IndividuumToPunkt(BergGeneration.BestesIndividuum);
  PunktAnzeigen(Punkt, BergGeneration.Fitness(BergGeneration.BestesIndividuum), clRed);
  for i:= 1 to StrToInt(E_Anzahl.Text)
  do PunktAnzeigen(BergGeneration.IndividuumToPunkt(BergGeneration.Individuum(i)),-1, clBlack);
end;

procedure TF_Berge.B_NeustartClick(Sender: TObject);
{ Neues Gebirge erstellen.
}
begin
  Gebirge.Free;
  Gebirge:= TGebirge.Create(PB_Gebirge);
  Gebirge.Anzeigen;
  B_Erste.SetFocus;
  E_Best.Text:= '';
end;

procedure TF_Berge.B_NaechsteClick(Sender: TObject);
{ Neue Generation erzeugen und anzeigen. Alte Generation stirbt komplett.
}
var Punkt: TPoint;
    i: integer;
begin
  E_Nr.Text:= IntToStr(StrToInt(E_Nr.Text)+1);
  BergGeneration.NaechsteGenerationErzeugen_BesteBleibenErhalten;
  BergGeneration.Anzeigen(PB_GenAlg);

  Gebirge.Anzeigen;

  Punkt:= BergGeneration.IndividuumToPunkt(BergGeneration.BestesIndividuum);
  PunktAnzeigen(Punkt, BergGeneration.Fitness(BergGeneration.BestesIndividuum), clRed);
  for i:= 1 to StrToInt(E_Anzahl.Text)
  do PunktAnzeigen(BergGeneration.IndividuumToPunkt(BergGeneration.Individuum(i)),-1, clBlack);
end;

procedure TF_Berge.B_BestClick(Sender: TObject);
{ Sucht iterativ den hchsten Punkt im Gebirge. Dies ist natrlich nur bei einem kleinen
  Gebirge wir in diesem Beispiel sinnvoll. Bei greren Gebirgen (n*m Punkte) wird die
  Laufzeit O(n*m) unakzeptabel.
}
var x,y: integer;
    Max, Maxx, Maxy: integer;
    d: integer;
begin
  Max:= 0;
  for x:= 1 to 255 do for y:= 1 to 255
  do begin
       d:= Gebirge.Punkte[x,y];
       if d>Max
       then begin
              Max:= d;
              Maxx:= x;
              Maxy:= y;
            end;
     end;
  E_Best.Text:= inttostr(Max);

  PunktAnzeigen(point(Maxx,Maxy), Max, clBlue);
end;

procedure TF_Berge.FormCreate(Sender: TObject);
begin
  timer1.enabled:= true;              { Erstellung des ersten Gebirges zeitverzgert starten, sonst Laufzeitfehler
                                        wegen MouseMove-Ereignis der Paintbox. }
end;

procedure TF_Berge.Timer1Timer(Sender: TObject);
begin
  timer1.enabled:= false;             { Hier wird nun das Gebirge erstellt. Zeitverzgerungs-Timer wird deaktiviert. }
  Gebirge:= TGebirge.Create(PB_Gebirge);
  Gebirge.Anzeigen;

end;

procedure TF_Berge.PB_GebirgeMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  if (x>0) and (Y>0) and (Gebirge<>NIL)
  then L_Hoehe.Caption:='Hhe: '+IntToStr(Gebirge.Punkte[x,y]); { Anzeige der aktuellen Gebirgshhe }

end;

end.
